/*
 * Copyright (C) 2014-2016 Brian L. Browning
 *
 * This file is part of Beagle
 *
 * Beagle is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Beagle is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package imp;

import blbutil.IntArray;

/**
 * <p>Class {@code HapToSeq} stores an array mapping haplotype indices to
 * allele sequences indices.
 * </p>
 * <p>Instances of class {@code HapToSeq} are immutable.
 * </p>
 *
 * @author Brian L. Browning {@code <browning@uw.edu>}
 */
public class HapToSeq {

    private final IntArray hap2Seq;
    private final int nSeq;

    /**
     * Constructs a new {@code HapAlleles} instance from the specified data.
     * The contract for this constructor and class is unspecified if
     * the sequence indices are not a subset of {@code 0, 1, ..., (nSeq - 1)}.
     *
     * @param hap2Seq an array mapping haplotype index to sequence index
     * @param nSeq the number of distinct sequence indices
     * @throws IllegalArgumentException if {@code nSeq < 1}
     * @throws NullPointerException if{@code hap2Seq == null}
     */
    public HapToSeq(int[] hap2Seq, int nSeq) {
        this.hap2Seq = IntArray.create(hap2Seq, 0, nSeq);
        this.nSeq = nSeq;
    }

    /**
     * Constructs a new {@code HapAlleles} instance from the specified data.
     * The contract for this constructor and class is unspecified if
     * the sequence indices are not a subset of {@code 0, 1, ..., (nSeq - 1)}.
     *
     * @param hap2Seq an array mapping haplotype index to sequence index
     * @param nSeq the number of distinct sequence indices
     * @throws NullPointerException if{@code hap2Seq == null}
     */
    public HapToSeq(IntArray hap2Seq, int nSeq) {
        this.hap2Seq = hap2Seq;
        this.nSeq = nSeq;
    }

    /**
     * Returns the number of haplotypes.
     * @return the number of haplotypes
     */
    public int nHaps() {
        return hap2Seq.size();
    }

    /**
     * Returns the number of sequences.
     * @return the number of sequences
     */
    public int nSeqs() {
        return nSeq;
    }

    /**
     * Returns the array mapping haplotype indices to sequence indices.
     * @return the array mapping haplotype indices to sequence indices
     */
    public IntArray hap2Seq() {
        return hap2Seq;
    }
}
