/*
 * Copyright (C) 2014-2016 Brian L. Browning
 *
 * This file is part of Beagle
 *
 * Beagle is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Beagle is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package vcf;

/**
 * <p>Interface {@code PhasedGT} represents phased, nonmissing genotype
 * data for a list of markers and a list of samples.
 * </p>
 * <p>
 * All instances of {@code PhasedGT} are required to be immutable.
 * </p>
 *
 * @author Brian L. Browning {@code <browning@uw.edu>}
 */
public interface PhasedGT extends GT {

    /**
     * Returns {@code true}.
     * @param marker the marker index
     * @param sample the sample index
     * @return {@code true}
     *
     * @throws IndexOutOfBoundsException if
     * {@code marker < 0 || marker >= this.nMarkers()}
     * @throws IndexOutOfBoundsException if
     * {@code sample < 0 || sample >= this.nSamples()}
     */
    @Override
    default boolean isPhased(int marker, int sample) {
        if (marker < 0 || marker >= this.nMarkers()) {
            throw new IndexOutOfBoundsException(String.valueOf(marker));
        }
        if (sample < 0 || sample >= this.nSamples()) {
            throw new IndexOutOfBoundsException(String.valueOf(sample));
        }
        return true;
    }

    /**
     * Returns {@code true}.
     * @param sample a sample index
     * @return {@code true}
     *
     * @throws IndexOutOfBoundsException if
     * {@code sample < 0 || sample >= this.nSamples()}
     */
    @Override
    default boolean isPhased(int sample) {
        if (sample < 0 || sample >= this.nSamples()) {
            throw new IndexOutOfBoundsException(String.valueOf(sample));
        }
        return true;
    }

    /**
     * Returns {@code true}.
     * @return {@code true}
     */
    @Override
    default boolean isPhased() {
        return true;
    }

    /**
     * Returns the probability that the specified marker allele is
     * present on the first haplotype of the specified sample.
     *
     * @param marker a marker index
     * @param sample a sample index
     * @param allele an allele index
     * @return the probability that the specified marker allele is
     * present on the first haplotype of the specified sample
     *
     * @throws IndexOutOfBoundsException if
     * {@code marker < 0 || marker >= this.nMarkers()}
     * @throws IndexOutOfBoundsException if
     * {@code sample < 0 || sample >= this.nSamples()}
     * @throws IndexOutOfBoundsException if
     * {@code allele < 0 || allele >= this.marker(marker).nAlleles()}
     *
     * @implSpec The default implementation returns {@code 1.0f} if
     * {@code allele1(marker, sample} == allele} and returns {@code 0.0f}
     * otherwise.
     */
    default float alProb1(int marker, int sample, int allele) {
        if (allele < 0 || allele >= this.marker(marker).nAlleles()) {
            throw new IndexOutOfBoundsException(String.valueOf(allele));
        }
        return allele==allele1(marker,sample) ? 1.0f : 0.0f;
    }

    /**
     * Returns the probability that the specified marker allele is
     * present on the second haplotype of the specified sample.
     *
     * @param marker a marker index
     * @param sample a sample index
     * @param allele an allele index
     * @return the probability that the specified marker allele is
     * present on the second haplotype of the specified sample.
     *
     * @throws IndexOutOfBoundsException if
     * {@code marker < 0 || marker >= this.nMarkers()}
     * @throws IndexOutOfBoundsException if
     * {@code sample < 0 || sample >= this.nSamples()}
     * @throws IndexOutOfBoundsException if
     * {@code allele < 0 || allele >= this.marker(marker).nAlleles()}
     *
     * @implSpec The default implementation returns {@code 1.0f} if
     * {@code allele2(marker, sample} == allele} and returns {@code 0.0f}
     * otherwise.
     */
    default float alProb2(int marker, int sample, int allele) {
        if (allele < 0 || allele >= this.marker(marker).nAlleles()) {
            throw new IndexOutOfBoundsException(String.valueOf(allele));
        }
        return allele==allele2(marker,sample) ? 1.0f : 0.0f;
    }

    /**
     * Returns the phased genotype probability, equal to
     * {@code (this.allele1(marker, sample, allele1)
     *        * this.allele2(marker, sample, allele2))}.
     *
     * @param marker a marker index
     * @param sample a sample index
     * @param allele1 allele index of the allele on the first haplotype
     * @param allele2 allele index of the allele on the second haplotype
     * @return the phased genotype probability equal to
     *       {@code (this.allele1(marker, sample, allele1)
     *              * this.allele2(marker, sample, allele2))}
     *
     * @throws IndexOutOfBoundsException if
     * {@code marker < 0 || marker >= this.nMarkers()}
     * @throws IndexOutOfBoundsException if
     * {@code sample < 0 || sample >= this.nSamples()}
     * @throws IndexOutOfBoundsException if
     * {@code allele1 < 0 || allele1 >= this.marker(marker).nAlleles()}
     * @throws IndexOutOfBoundsException if
     * {@code allele2 < 0 || allele2 >= this.marker(marker).nAlleles()}
     */
    default float gtProb(int marker, int sample, int allele1, int allele2) {
        return alProb1(marker, sample, allele1)*alProb2(marker, sample, allele2);
    }
}
