/*
 * Copyright (C) 2014-2016 Brian L. Browning
 *
 * This file is part of Beagle
 *
 * Beagle is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Beagle is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package vcf;

import beagleutil.Samples;
import blbutil.IntArray;
import java.util.Arrays;

/**
 * <p>Interface {@code GTRec} represents represents genotype data for one
 * marker.
 * </p>
 * <p>All instances of {@code GTRec} are required to be immutable.
 * </p>
 *
 * @author Brian L. Browning {@code <browning@uw.edu>}
 */
public interface RefGTRec extends GTRec {

    /**
     * Returns an allele-coded {@code RefGTRec} instance for the
     * specified data.
     * @param rec the phased, non-missing genotype data
     * @return an allele-coded {@code RefGTRec} instance for the
     * specified data
     * @throws NullPointerException if {@code rec == null}
     */
    static RefGTRec alleleCodedInstance(RefGTRec rec) {
        if (rec.isAlleleCoded()) {
            return rec;
        }
        if (rec.marker().nAlleles()==2) {
            return new LowMafRefDiallelicGT(rec);
        }
        else {
            return new LowMafRefGT(rec);
        }
    }

    /**
     * Constructs and returns a new allele-coded {@code RefGTRec} instance
     * from the specified data.
     *
     * @param gtp a VCF record parser that extracts sample genotypes
     * @return an allele-coded {@code RefGTRec} instance
     *
     * @throws IllegalArgumentException if the VCF record contains an
     * unphased genotype or missing allele
     * @throws IllegalArgumentException if a format error is detected in the
     * VCF record
     * @throws NullPointerException if {@code gtp == null}
     */
    static RefGTRec alleleCodedInstance(VcfRecGTParser gtp) {
        if (gtp.nAlleles()==2) {
            return new LowMafRefDiallelicGT(gtp);
        }
        else {
            return new LowMafRefGT(gtp);
        }
    }

    /**
     * Constructs and returns a new allele-coded {@code RefGTRec} instance
     * from the specified data. The contract for this class is unspecified
     * if a haplotype index is duplicated in the specified {@code hapIndices}
     * array.
     *
     * @param marker the marker
     * @param samples the samples
     * @param hapIndices an array whose {@code j}-th element is {@code null}
     * if {@code j} is the major allele with lowest index, and otherwise is
     * an array of indices of haplotypes that carry the {@code j}-th allele
     * sorted in increasing order
     * @return an allele-coded {@code RefGTRec} instance
     *
     * @throws IllegalArgumentException if the {@code (hapIndices[j] == null)}
     * and {@code j} is not the major allele with lowest index, or if
     * {@code (hapIndices[j] != null)} and {@code j} is the major allele with
     * lowest index
     * @throws IllegalArgumentException if any element of {@code hapIndices}
     * is not not a sorted array of distinct haplotype indices
     * between 0 (inclusive) and {@code 2*samples.nSamples()} (exclusive)
     * @throws IllegalArgumentException if
     * {@code marker.nAlleles() != hapIndices.length}
     * @throws NullPointerException if
     * {@code marker == null || samples == null || hapIndices == null}
     */
    static RefGTRec alleleCodedInstance(Marker marker, Samples samples,
            int[][] hapIndices) {
        if (marker.nAlleles()==2) {
            return new LowMafRefDiallelicGT(marker, samples, hapIndices);
        }
        else {
            return new LowMafRefGT(marker, samples, hapIndices);
        }
    }

    /**
     * Returns an array whose {@code j}-th element is {@code null}
     * if {@code j} is the major allele with lowest index, and otherwise is
     * an array of indices of haplotypes that carry the {@code j}-th allele
     * sorted in increasing order
     * @param rec the phased, non-missing genotype data
     * @return an array whose {@code j}-th element is {@code null}
     * if {@code j} is the major allele with lowest index, and otherwise is
     * an array of indices of haplotypes that carry the {@code j}-th allele
     * sorted in increasing order
     * @throws NullPointerException if {@code rec == null}
     */
    static int[][] hapIndices(RefGTRec rec) {
        int[] alCnts = new int[rec.marker().nAlleles()];
        for (int h=0, n=rec.nHaps(); h<n; ++h) {
            ++alCnts[rec.allele(h)];
        }
        int majAllele = 0;
        for (int al=1; al<alCnts.length; ++al) {
            if (alCnts[al] > alCnts[majAllele]) {
                majAllele = al;
            }
        }
        int[][] hapIndices = new int[alCnts.length][];
        for (int al=0; al<alCnts.length; ++al) {
            if (al!=majAllele) {
                hapIndices[al] = new int[alCnts[al]];
            }
        }
        Arrays.fill(alCnts, 0);
        for (int h=0, n=rec.nHaps(); h<n; ++h) {
            int al = rec.allele(h);
            if (al!=majAllele) {
                hapIndices[al][alCnts[al]++] = h;
            }
        }
        return hapIndices;
    }

    /**
     * Returns {@code true}.
     * @param sample the sample index
     * @return {@code true}
     *
     * @throws IndexOutOfBoundsException if
     * {@code sample < 0 || sample >= this.nSamples()}
     */
    @Override
    default boolean isPhased(int sample) {
        if (sample < 0 || sample >= this.nSamples()) {
            throw new IndexOutOfBoundsException(String.valueOf(sample));
        }
        return true;
    }

    /**
     * Returns {@code true}.
     * @return {@code true}
     */
    @Override
    default boolean isPhased() {
        return true;
    }

    /**
     * Returns a {@code RefGTRec} instance with the specified marker and the
     * same samples and haplotype alleles as {@code this}.
     * @param marker a marker
     * @return a {@code RefGTRec} instance with the specified marker and the
     * same samples and haplotype alleles as {@code this}
     * @throws IllegalArgumentException if
     * {@code marker.nAlleles() < this.nAlleles()}
     * @throws NullPointerException if {@code marker == null}
     */
    RefGTRec changeMarker(Marker marker);

    /**
     * Returns {@code true} if this instance stores the indices of haplotypes
     * that carry non-major alleles, and returns {@code false} otherwise.
     * The default implementation returns {@code false}
     *
     * @return {@code true} if this instance stores the indices of haplotypes
     * that carry non-major alleles
     *
     * @implNote The default implementation returns {@code true}
     */
    default boolean isAlleleCoded() {
        return false;
    }

    /**
     * Returns the index of the major allele.
     * @return the index of the major allele
     * @throws UnsupportedOperationException if
     * {@code isAlleleCoded() == false}
     *
     * @implNote The default implementation throws an
     * {@code UnsupportedOperationException}
     */
    default int majorAllele() {
        throw new UnsupportedOperationException();
    }

    /**
     * Returns the number of haplotypes that carry the specified allele.
     * The default implementation throws an {@code UnsupportedOperationException}
     * @param allele an allele index
     * @return the number of haplotypes that carry the specified allele
     * @throws IllegalArgumentException if
     * {@code allele == this.majorAllele()}
     * @throws IndexOutOfBoundsException if
     * {@code allele < 0 ||  allele >= this.nAlleles()}
     * @throws UnsupportedOperationException if
     * {@code isAlleleCoded() == false}
     *
     * @implNote The default implementation throws an
     * {@code UnsupportedOperationException}
     */
    default int alleleCount(int allele) {
        throw new UnsupportedOperationException();
    }

    /**
     * Returns index of the haplotype that carries the specified copy of the
     * specified allele.
     * The default implementation throws an {@code UnsupportedOperationException}
     * @param allele an allele index
     * @param copy a copy index
     * @return index of the haplotype that carries the specified allele
     * @throws IllegalArgumentException if
     * {@code allele == this.majorAllele()}
     * @throws IndexOutOfBoundsException if
     * {@code allele < 0 ||  allele >= this.nAlleles()}
     * @throws IndexOutOfBoundsException if
     * {@code copy < 0 ||  copy >= this.alleleCount(allele)}
     * @throws UnsupportedOperationException if
     * {@code isAlleleCoded() == false}
     *
     * @implNote The default implementation throws an
     * {@code UnsupportedOperationException}
     */
    default int hapIndex(int allele, int copy) {
        throw new UnsupportedOperationException();
    }

    /**
     * Returns {@code true} if the specified haplotype carries the specified
     * allele and return {@code false} otherwise.
     * @param allele an allele index
     * @param hap a haplotype index
     * @return {@code true} if the specified haplotype carries the specified
     * allele
     * @throws IndexOutOfBoundsException if
     * {@code hap < 0 || hap >= this.nHaps()}
     * @throws IndexOutOfBoundsException if
     * {@code allele < 0 || allele >= this.nAlleles()}
     */
    default boolean isCarrier(int allele, int hap) {
        return allele(hap)==allele;
    }

    /**
     * Returns an {@code IntArray} mapping haplotype indices to
     * sequences.
     *
     * @return the sequences for each haplotype
     * @throws UnsupportedOperationException if
     * {@code isAlleleCoded() == true}
     *
     * @implNote The default implementation throws an
     * {@code UnsupportedOperationException}
     */
    default IntArray hapToSeq() {
        throw new UnsupportedOperationException();
    }

    /**
     * Returns an {@code IntArray} mapping the sequence indices
     * to alleles carried by the sequence for this record
     *
     * @return the alleles for each sequence
     * @throws UnsupportedOperationException if
     * {@code isAlleleCoded() == true}
     *
     * @implNote The default implementation throws an
     * {@code UnsupportedOperationException}
     */
    default IntArray seqToAllele() {
        throw new UnsupportedOperationException();
    }
}